@@ -1,6 +1,7 @@
-/* libp11, a simple layer on to of PKCS#11 API
+/* libp11, a simple layer on top of PKCS#11 API
  * Copyright (C) 2017 Douglas E. Engert <deengert@gmail.com>
- * Copyright (C) 2017-2018 Michał Trojnara <Michal.Trojnara@stunnel.org>
+ * Copyright (C) 2017-2025 Michał Trojnara <Michal.Trojnara@stunnel.org>
+ * Copyright © 2025 Mobi - Com Polska Sp. z o.o.
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -20,6 +21,8 @@
 #include "libp11-int.h"
 #include <string.h>
 
+#if OPENSSL_VERSION_NUMBER < 0x40000000L
+
 static int (*orig_pkey_rsa_sign_init) (EVP_PKEY_CTX *ctx);
 static int (*orig_pkey_rsa_sign) (EVP_PKEY_CTX *ctx,
 	unsigned char *sig, size_t *siglen,
@@ -29,12 +32,22 @@ static int (*orig_pkey_rsa_decrypt) (EVP_PKEY_CTX *ctx,
 	unsigned char *out, size_t *outlen,
 	const unsigned char *in, size_t inlen);
 
-#ifndef OPENSSL_NO_EC
+# ifndef OPENSSL_NO_EC
 static int (*orig_pkey_ec_sign_init) (EVP_PKEY_CTX *ctx);
 static int (*orig_pkey_ec_sign) (EVP_PKEY_CTX *ctx,
 	unsigned char *sig, size_t *siglen,
 	const unsigned char *tbs, size_t tbslen);
-#endif /* OPENSSL_NO_EC */
+
+#  if OPENSSL_VERSION_NUMBER >= 0x30000000L
+static int (*orig_pkey_ed25519_digestsign)(EVP_MD_CTX *ctx,
+	unsigned char *sig, size_t *siglen,
+	const unsigned char *tbs, size_t tbslen);
+static int (*orig_pkey_ed448_digestsign)(EVP_MD_CTX *ctx,
+	unsigned char *sig, size_t *siglen,
+	const unsigned char *tbs, size_t tbslen);
+#  endif /* OPENSSL_VERSION_NUMBER >= 0x30000000L */
+# endif /* OPENSSL_NO_EC */
+#endif /* OPENSSL_VERSION_NUMBER < 0x40000000L */
 
 #if OPENSSL_VERSION_NUMBER < 0x10002000L || defined(LIBRESSL_VERSION_NUMBER)
 
@@ -141,6 +154,7 @@ static void EVP_PKEY_meth_get_decrypt(EVP_PKEY_METHOD *pmeth,
 }
 #endif
 
+#if OPENSSL_VERSION_NUMBER < 0x40000000L
 static CK_MECHANISM_TYPE pkcs11_md2ckm(const EVP_MD *md)
 {
 	switch (EVP_MD_type(md)) {
@@ -198,7 +212,7 @@ static CK_RSA_PKCS_MGF_TYPE pkcs11_md2ckg(const EVP_MD *md)
 }
 
 static int pkcs11_params_pss(CK_RSA_PKCS_PSS_PARAMS *pss,
-		EVP_PKEY_CTX *ctx)
+		EVP_PKEY_CTX *ctx, PKCS11_CTX_private *pctx)
 {
 	const EVP_MD *sig_md, *mgf1_md;
 	EVP_PKEY *evp_pkey;
@@ -216,6 +230,8 @@ static int pkcs11_params_pss(CK_RSA_PKCS_PSS_PARAMS *pss,
 		salt_len = EVP_MD_size(sig_md);
 		break;
 	case -2:
+		if (!ctx)
+			return -1;
 		evp_pkey = EVP_PKEY_CTX_get0_pkey(ctx);
 		if (!evp_pkey)
 			return -1;
@@ -225,10 +241,8 @@ static int pkcs11_params_pss(CK_RSA_PKCS_PSS_PARAMS *pss,
 		if (salt_len < 0) /* integer underflow detected */
 			return -1;
 	}
-#ifdef DEBUG
-	fprintf(stderr, "salt_len=%d sig_md=%s mdf1_md=%s\n",
+	pkcs11_log(pctx, LOG_DEBUG, "salt_len=%d sig_md=%s mdf1_md=%s\n",
 		salt_len, EVP_MD_name(sig_md), EVP_MD_name(mgf1_md));
-#endif
 
 	/* fill the CK_RSA_PKCS_PSS_PARAMS structure */
 	memset(pss, 0, sizeof(CK_RSA_PKCS_PSS_PARAMS));
@@ -241,7 +255,7 @@ static int pkcs11_params_pss(CK_RSA_PKCS_PSS_PARAMS *pss,
 }
 
 static int pkcs11_params_oaep(CK_RSA_PKCS_OAEP_PARAMS *oaep,
-		EVP_PKEY_CTX *ctx)
+		EVP_PKEY_CTX *ctx, PKCS11_CTX_private *pctx)
 {
 	const EVP_MD *oaep_md, *mgf1_md;
 
@@ -250,10 +264,9 @@ static int pkcs11_params_oaep(CK_RSA_PKCS_OAEP_PARAMS *oaep,
 		return -1;
 	if (EVP_PKEY_CTX_get_rsa_mgf1_md(ctx, &mgf1_md) <= 0)
 		return -1;
-#ifdef DEBUG
-	fprintf(stderr, "oaep_md=%s mdf1_md=%s\n",
+
+	pkcs11_log(pctx, LOG_DEBUG, "oaep_md=%s mdf1_md=%s\n",
 		EVP_MD_name(oaep_md), EVP_MD_name(mgf1_md));
-#endif
 
 	/* fill the CK_RSA_PKCS_OAEP_PARAMS structure */
 	memset(oaep, 0, sizeof(CK_RSA_PKCS_OAEP_PARAMS));
@@ -275,7 +288,7 @@ static int pkcs11_try_pkey_rsa_sign(EVP_PKEY_CTX *evp_pkey_ctx,
 	EVP_PKEY *pkey;
 	RSA *rsa;
 	int rv = 0, padding;
-	CK_ULONG size = *siglen;
+	CK_ULONG size = (CK_ULONG)*siglen;
 	PKCS11_OBJECT_private *key;
 	PKCS11_SLOT_private *slot;
 	PKCS11_CTX_private *ctx;
@@ -284,14 +297,11 @@ static int pkcs11_try_pkey_rsa_sign(EVP_PKEY_CTX *evp_pkey_ctx,
 	CK_MECHANISM mechanism;
 	CK_RSA_PKCS_PSS_PARAMS pss_params;
 
-#ifdef DEBUG
-	fprintf(stderr, "%s:%d pkcs11_try_pkey_rsa_sign() "
-		"sig=%p *siglen=%lu tbs=%p tbslen=%lu\n",
-		__FILE__, __LINE__, sig, *siglen, tbs, tbslen);
-#endif
 	/* RSA method has EVP_PKEY_FLAG_AUTOARGLEN set. OpenSSL core will handle
 	 * the size inquiry internally. */
-	 if (!sig)
+	if (!sig)
+		return -1;
+	if (!evp_pkey_ctx)
 		return -1;
 
 	pkey = EVP_PKEY_CTX_get0_pkey(evp_pkey_ctx);
@@ -307,8 +317,13 @@ static int pkcs11_try_pkey_rsa_sign(EVP_PKEY_CTX *evp_pkey_ctx,
 
 	slot = key->slot;
 	ctx = slot->ctx;
-	if (!evp_pkey_ctx)
+	if (!ctx)
 		return -1;
+#ifdef DEBUG
+	pkcs11_log(ctx, LOG_DEBUG, "%s:%d pkcs11_try_pkey_rsa_sign() "
+		"sig=%p *siglen=%lu tbs=%p tbslen=%lu\n",
+		__FILE__, __LINE__, sig, *siglen, tbs, tbslen);
+#endif
 	if (EVP_PKEY_CTX_get_signature_md(evp_pkey_ctx, &sig_md) <= 0)
 		return -1;
 	if (tbslen != (size_t)EVP_MD_size(sig_md))
@@ -318,21 +333,16 @@ static int pkcs11_try_pkey_rsa_sign(EVP_PKEY_CTX *evp_pkey_ctx,
 	EVP_PKEY_CTX_get_rsa_padding(evp_pkey_ctx, &padding);
 	switch (padding) {
 	case RSA_PKCS1_PSS_PADDING:
-#ifdef DEBUG
-		fprintf(stderr, "%s:%d padding=RSA_PKCS1_PSS_PADDING\n",
-			__FILE__, __LINE__);
-#endif
-		if (pkcs11_params_pss(&pss_params, evp_pkey_ctx) < 0)
+		pkcs11_log(ctx, LOG_DEBUG, "padding=RSA_PKCS1_PSS_PADDING\n");
+		if (pkcs11_params_pss(&pss_params, evp_pkey_ctx, ctx) < 0)
 			return -1;
 		mechanism.mechanism = CKM_RSA_PKCS_PSS;
 		mechanism.pParameter = &pss_params;
 		mechanism.ulParameterLen = sizeof pss_params;
 		break;
 	default:
-#ifdef DEBUG
-		fprintf(stderr, "%s:%d unsupported padding: %d\n",
+		pkcs11_log(ctx, LOG_DEBUG, "%s:%d unsupported padding: %d\n",
 			__FILE__, __LINE__, padding);
-#endif
 		return -1;
 	} /* end switch(padding) */
 
@@ -341,16 +351,20 @@ static int pkcs11_try_pkey_rsa_sign(EVP_PKEY_CTX *evp_pkey_ctx,
 
 	rv = CRYPTOKI_call(ctx,
 		C_SignInit(session, &mechanism, key->object));
-	if (!rv && key->always_authenticate == CK_TRUE)
+	if (rv != CKR_OK) {
+		pkcs11_log(ctx, LOG_DEBUG, "%s:%d C_SignInit rv=%d\n",
+			__FILE__, __LINE__, rv);
+	} else if (key->always_authenticate == CK_TRUE)
 		rv = pkcs11_authenticate(key, session);
-	if (!rv)
+	if (rv == CKR_OK) {
 		rv = CRYPTOKI_call(ctx,
-			C_Sign(session, (CK_BYTE_PTR)tbs, tbslen, sig, &size));
+			C_Sign(session, (CK_BYTE_PTR)tbs, (CK_ULONG)tbslen, sig, &size));
+		if (rv != CKR_OK) {
+			pkcs11_log(ctx, LOG_DEBUG, "%s:%d C_Sign rv=%d\n",
+				__FILE__, __LINE__, rv);
+		}
+	}
 	pkcs11_put_session(slot, session);
-#ifdef DEBUG
-	fprintf(stderr, "%s:%d C_SignInit or C_Sign rv=%d\n",
-		__FILE__, __LINE__, rv);
-#endif
 
 	if (rv != CKR_OK)
 		return -1;
@@ -377,7 +391,7 @@ static int pkcs11_try_pkey_rsa_decrypt(EVP_PKEY_CTX *evp_pkey_ctx,
 	EVP_PKEY *pkey;
 	RSA *rsa;
 	int rv = 0, padding;
-	CK_ULONG size = *outlen;
+	CK_ULONG size = (CK_ULONG)*outlen;
 	PKCS11_OBJECT_private *key;
 	PKCS11_SLOT_private *slot;
 	PKCS11_CTX_private *ctx;
@@ -385,14 +399,11 @@ static int pkcs11_try_pkey_rsa_decrypt(EVP_PKEY_CTX *evp_pkey_ctx,
 	CK_MECHANISM mechanism;
 	CK_RSA_PKCS_OAEP_PARAMS oaep_params;
 
-#ifdef DEBUG
-	fprintf(stderr, "%s:%d pkcs11_try_pkey_rsa_decrypt() "
-		"out=%p *outlen=%lu in=%p inlen=%lu\n",
-		__FILE__, __LINE__, out, *outlen, in, inlen);
-#endif
 	/* RSA method has EVP_PKEY_FLAG_AUTOARGLEN set. OpenSSL core will handle
 	 * the size inquiry internally. */
-	 if (!out)
+	if (!out)
+		return -1;
+	if (!evp_pkey_ctx)
 		return -1;
 
 	pkey = EVP_PKEY_CTX_get0_pkey(evp_pkey_ctx);
@@ -408,38 +419,33 @@ static int pkcs11_try_pkey_rsa_decrypt(EVP_PKEY_CTX *evp_pkey_ctx,
 
 	slot = key->slot;
 	ctx = slot->ctx;
-
-	if (!evp_pkey_ctx)
+	if (!ctx)
 		return -1;
-
+#ifdef DEBUG
+	pkcs11_log(ctx, LOG_DEBUG, "%s:%d pkcs11_try_pkey_rsa_decrypt() "
+		"out=%p *outlen=%lu in=%p inlen=%lu\n",
+		__FILE__, __LINE__, out, *outlen, in, inlen);
+#endif
 	memset(&mechanism, 0, sizeof mechanism);
 	EVP_PKEY_CTX_get_rsa_padding(evp_pkey_ctx, &padding);
 	switch (padding) {
 	case RSA_PKCS1_OAEP_PADDING:
-#ifdef DEBUG
-		fprintf(stderr, "%s:%d padding=RSA_PKCS1_OAEP_PADDING\n",
-			__FILE__, __LINE__);
-#endif
-		if (pkcs11_params_oaep(&oaep_params, evp_pkey_ctx) < 0)
+		pkcs11_log(ctx, LOG_DEBUG, "padding=RSA_PKCS1_OAEP_PADDING\n");
+		if (pkcs11_params_oaep(&oaep_params, evp_pkey_ctx, ctx) < 0)
 			return -1;
 		mechanism.mechanism = CKM_RSA_PKCS_OAEP;
 		mechanism.pParameter = &oaep_params;
 		mechanism.ulParameterLen = sizeof oaep_params;
 		break;
 	case RSA_PKCS1_PADDING:
-#ifdef DEBUG
-		fprintf(stderr, "%s:%d padding=RSA_PKCS1_PADDING\n",
-			__FILE__, __LINE__);
-#endif
+		pkcs11_log(ctx, LOG_DEBUG, "padding=RSA_PKCS1_PADDING\n");
 		mechanism.mechanism = CKM_RSA_PKCS;
 		mechanism.pParameter = NULL;
 		mechanism.ulParameterLen = 0;
 		break;
 	default:
-#ifdef DEBUG
-		fprintf(stderr, "%s:%d unsupported padding: %d\n",
+		pkcs11_log(ctx, LOG_DEBUG, "%s:%d unsupported padding: %d\n",
 			__FILE__, __LINE__, padding);
-#endif
 		return -1;
 	} /* end switch(padding) */
 
@@ -448,16 +454,20 @@ static int pkcs11_try_pkey_rsa_decrypt(EVP_PKEY_CTX *evp_pkey_ctx,
 
 	rv = CRYPTOKI_call(ctx,
 		C_DecryptInit(session, &mechanism, key->object));
-	if (!rv && key->always_authenticate == CK_TRUE)
+	if (rv != CKR_OK) {
+		pkcs11_log(ctx, LOG_DEBUG, "%s:%d C_DecryptInit rv=%d\n",
+			__FILE__, __LINE__, rv);
+	} else if (key->always_authenticate == CK_TRUE)
 		rv = pkcs11_authenticate(key, session);
-	if (!rv)
+	if (rv == CKR_OK) {
 		rv = CRYPTOKI_call(ctx,
-			C_Decrypt(session, (CK_BYTE_PTR)in, inlen, out, &size));
+			C_Decrypt(session, (CK_BYTE_PTR)in, (CK_ULONG)inlen, out, &size));
+		if (rv != CKR_OK) {
+			pkcs11_log(ctx, LOG_DEBUG, "%s:%d C_Decrypt rv=%d\n",
+				__FILE__, __LINE__, rv);
+		}
+	}
 	pkcs11_put_session(slot, session);
-#ifdef DEBUG
-	fprintf(stderr, "%s:%d C_DecryptInit or C_Decrypt rv=%d\n",
-		__FILE__, __LINE__, rv);
-#endif
 
 	if (rv != CKR_OK)
 		return -1;
@@ -477,7 +487,7 @@ static int pkcs11_pkey_rsa_decrypt(EVP_PKEY_CTX *evp_pkey_ctx,
 	return ret;
 }
 
-static EVP_PKEY_METHOD *pkcs11_pkey_method_rsa()
+static EVP_PKEY_METHOD *pkcs11_pkey_method_rsa(void)
 {
 	EVP_PKEY_METHOD *orig_meth, *new_meth;
 
@@ -510,7 +520,7 @@ static int pkcs11_try_pkey_ec_sign(EVP_PKEY_CTX *evp_pkey_ctx,
 	EVP_PKEY *pkey;
 	EC_KEY *eckey;
 	int rv = CKR_GENERAL_ERROR;
-	CK_ULONG size = *siglen;
+	CK_ULONG size = (CK_ULONG)*siglen;
 	PKCS11_OBJECT_private *key;
 	PKCS11_SLOT_private *slot;
 	PKCS11_CTX_private *ctx;
@@ -519,15 +529,11 @@ static int pkcs11_try_pkey_ec_sign(EVP_PKEY_CTX *evp_pkey_ctx,
 	ECDSA_SIG *ossl_sig;
 	CK_MECHANISM mechanism;
 
-#ifdef DEBUG
-	fprintf(stderr, "%s:%d pkcs11_try_pkey_ec_sign() "
-		"sig=%p *siglen=%lu tbs=%p tbslen=%lu\n",
-		__FILE__, __LINE__, sig, *siglen, tbs, tbslen);
-#endif
-
 	ossl_sig = ECDSA_SIG_new();
 	if (!ossl_sig)
 		goto error;
+	if (!evp_pkey_ctx)
+		goto error;
 
 	pkey = EVP_PKEY_CTX_get0_pkey(evp_pkey_ctx);
 	if (!pkey)
@@ -552,10 +558,13 @@ static int pkcs11_try_pkey_ec_sign(EVP_PKEY_CTX *evp_pkey_ctx,
 
 	slot = key->slot;
 	ctx = slot->ctx;
-
-	if (!evp_pkey_ctx)
+	if (!ctx)
 		goto error;
-
+#ifdef DEBUG
+	pkcs11_log(ctx, LOG_DEBUG, "%s:%d pkcs11_try_pkey_ec_sign() "
+		"sig=%p *siglen=%lu tbs=%p tbslen=%lu\n",
+		__FILE__, __LINE__, sig, *siglen, tbs, tbslen);
+#endif
 	if (EVP_PKEY_CTX_get_signature_md(evp_pkey_ctx, &sig_md) <= 0)
 		goto error;
 
@@ -570,18 +579,21 @@ static int pkcs11_try_pkey_ec_sign(EVP_PKEY_CTX *evp_pkey_ctx,
 		return -1;
 	rv = CRYPTOKI_call(ctx,
 		C_SignInit(session, &mechanism, key->object));
-	if (!rv && key->always_authenticate == CK_TRUE)
+	if (rv != CKR_OK) {
+		pkcs11_log(ctx, LOG_DEBUG, "%s:%d C_SignInit rv=%d\n",
+			__FILE__, __LINE__, rv);
+	} else if (key->always_authenticate == CK_TRUE)
 		rv = pkcs11_authenticate(key, session);
-	if (!rv)
+	if (rv == CKR_OK) {
 		rv = CRYPTOKI_call(ctx,
-			C_Sign(session, (CK_BYTE_PTR)tbs, tbslen, sig, &size));
+			C_Sign(session, (CK_BYTE_PTR)tbs, (CK_ULONG)tbslen, sig, &size));
+		if (rv != CKR_OK) {
+			pkcs11_log(ctx, LOG_DEBUG, "%s:%d C_Sign rv=%d\n",
+				__FILE__, __LINE__, rv);
+		}
+	}
 	pkcs11_put_session(slot, session);
 
-#ifdef DEBUG
-	fprintf(stderr, "%s:%d C_SignInit or C_Sign rv=%d\n",
-		__FILE__, __LINE__, rv);
-#endif
-
 	if (rv == CKR_OK) {
 		BIGNUM *r = BN_bin2bn(sig, size/2, NULL);
 		BIGNUM *s = BN_bin2bn(sig + size/2, size/2, NULL);
@@ -606,6 +618,179 @@ error:
 	return 1;
 }
 
+# if OPENSSL_VERSION_NUMBER >= 0x30000000L
+/* PKCS#11 sign implementation for Ed25519 / Ed448 */
+static int pkcs11_eddsa_sign(unsigned char *sigret, unsigned int *siglen,
+	const unsigned char *tbs, unsigned int tbslen, PKCS11_OBJECT_private *key)
+{
+	int rv;
+	PKCS11_SLOT_private *slot = key->slot;
+	PKCS11_CTX_private *ctx = slot->ctx;
+	CK_SESSION_HANDLE session;
+	CK_MECHANISM mechanism;
+	CK_ULONG ck_siglen = (CK_ULONG)(*siglen);
+	CK_ULONG ck_tbslen = (CK_ULONG)tbslen;
+
+	if (!ctx)
+		return -1;
+
+	/* PureEdDSA, no prehash */
+	memset(&mechanism, 0, sizeof(mechanism));
+	mechanism.mechanism = CKM_EDDSA;
+
+#ifdef DEBUG
+	pkcs11_log(ctx, LOG_DEBUG, "%s:%d pkcs11_eddsa_sign() "
+		"sigret=%p *siglen=%u tbs=%p tbslen=%u\n",
+		__FILE__, __LINE__, sigret, *siglen, tbs, tbslen);
+#endif
+	if (pkcs11_get_session(slot, 0, &session))
+		return -1;
+
+	rv = CRYPTOKI_call(ctx,
+		C_SignInit(session, &mechanism, key->object));
+	if (!rv && key->always_authenticate == CK_TRUE)
+		rv = pkcs11_authenticate(key, session);
+	if (!rv)
+		rv = CRYPTOKI_call(ctx,
+			C_Sign(session, (CK_BYTE_PTR)tbs, ck_tbslen, sigret, &ck_siglen));
+	pkcs11_put_session(slot, session);
+
+	if (rv) {
+		CKRerr(CKR_F_PKCS11_EDDSA_SIGN, rv);
+		return -1;
+	}
+	*siglen = (unsigned int)ck_siglen;
+	return (int)ck_siglen;
+}
+
+/*
+ * EVP_PKEY method sign wrapper for EdDSA.
+ * This function is invoked internally by EVP_PKEY_sign().
+ * If the key belongs to PKCS#11, perform signing via pkcs11_eddsa_sign().
+ */
+static int pkcs11_eddsa_pmeth_sign(EVP_PKEY_CTX *evp_pkey_ctx,
+	unsigned char *sig, size_t *siglen,
+	const unsigned char *tbs, size_t tbslen)
+{
+	EVP_PKEY *pkey;
+	PKCS11_OBJECT_private *key;
+	unsigned int tmp_len;
+	int rv;
+
+	if (!evp_pkey_ctx)
+		return -1;
+
+	if (*siglen > UINT_MAX)
+		return 0;
+
+	pkey = EVP_PKEY_CTX_get0_pkey(evp_pkey_ctx);
+	if (!pkey)
+		return -1;
+
+	key = pkcs11_get_ex_data_pkey(pkey);
+	if (!key)
+		return -1;
+
+	if (check_object_fork(key) < 0)
+		return -1;
+
+	tmp_len = (unsigned int)*siglen;
+	rv = pkcs11_eddsa_sign(sig, &tmp_len, tbs, (unsigned int)tbslen, key);
+	if (rv < 0)
+		return -1;
+
+	*siglen = tmp_len;
+	return 1;
+}
+
+/*
+ * Custom EVP_PKEY_METHOD digestsign implementation for EdDSA (Ed25519/Ed448)
+ *
+ * This function supports the two-step signing process used by EVP_DigestSign*():
+ *   1. Query the required signature length (sig == NULL).
+ *   2. Perform the actual signing when a buffer is provided (sig != NULL).
+ *
+ * If the key is managed by PKCS#11, the signing is performed via pkcs11_eddsa_sign().
+ * Otherwise, the call is delegated to the original OpenSSL Ed25519/Ed448 implementation.
+ */
+static int pkcs11_eddsa_pmeth_digestsign(EVP_MD_CTX *ctx, unsigned char *sig,
+	size_t *siglen, const unsigned char *tbs, size_t tbslen)
+{
+	EVP_PKEY *pkey;
+	PKCS11_OBJECT_private *key;
+	unsigned int tmp_len;
+	int rv;
+
+	pkey = EVP_PKEY_CTX_get0_pkey(EVP_MD_CTX_pkey_ctx(ctx));
+	if (!pkey)
+		return -1;
+
+	key = pkcs11_get_ex_data_pkey(pkey);
+	if (!key)
+		return -1;
+
+	/* Step 1: caller asks for signature length only */
+	if (sig == NULL) {
+		if (EVP_PKEY_id(pkey) == EVP_PKEY_ED25519)
+			*siglen = 64; /* fixed size for Ed25519 */
+		else if (EVP_PKEY_id(pkey) == EVP_PKEY_ED448)
+			*siglen = 114; /* fixed size for Ed448 */
+		else
+			return -1;
+		/* success: report the expected signature length only,
+		 * no signing is performed in this call */
+		return 1;
+	}
+
+	/* Step 2: actual signing */
+	tmp_len = (unsigned int)*siglen;
+	rv = pkcs11_eddsa_sign(sig, &tmp_len, tbs, (unsigned int)tbslen, key);
+	if (rv < 0)
+		return 1;
+
+	*siglen = tmp_len;
+	return 1;
+}
+
+static int pkcs11_pkey_ed25519_digestsign(EVP_MD_CTX *ctx,
+		unsigned char *sig, size_t *siglen,
+		const unsigned char *tbs, size_t tbslen)
+{
+	int ret;
+
+	ret = pkcs11_eddsa_pmeth_digestsign(ctx, sig, siglen, tbs, tbslen);
+	if (ret < 0)
+		ret = (*orig_pkey_ed25519_digestsign)(ctx, sig, siglen, tbs, tbslen);
+	return ret;
+}
+
+static int pkcs11_pkey_ed448_digestsign(EVP_MD_CTX *ctx,
+		unsigned char *sig, size_t *siglen,
+		const unsigned char *tbs, size_t tbslen)
+{
+	int ret;
+
+	ret = pkcs11_eddsa_pmeth_digestsign(ctx, sig, siglen, tbs, tbslen);
+	if (ret < 0)
+		ret = (*orig_pkey_ed448_digestsign)(ctx, sig, siglen, tbs, tbslen);
+	return ret;
+}
+
+static int pkcs11_eddsa_pmeth_ctrl(EVP_PKEY_CTX *ctx, int type, int p1, void *p2)
+{
+	(void)ctx;
+	(void)p1;
+	switch (type) {
+	case EVP_PKEY_CTRL_MD:
+		if (p2 == NULL)
+			return 1; /* Accept NULL digest */
+		return 0; /* Reject if caller tries to set a digest */
+	default:
+		return -2; /* command not supported */
+	}
+}
+# endif /* OPENSSL_VERSION_NUMBER >= 0x30000000L */
+
 static int pkcs11_pkey_ec_sign(EVP_PKEY_CTX *evp_pkey_ctx,
 		unsigned char *sig, size_t *siglen,
 		const unsigned char *tbs, size_t tbslen)
@@ -618,7 +803,7 @@ static int pkcs11_pkey_ec_sign(EVP_PKEY_CTX *evp_pkey_ctx,
 	return ret;
 }
 
-static EVP_PKEY_METHOD *pkcs11_pkey_method_ec()
+static EVP_PKEY_METHOD *pkcs11_pkey_method_ec(void)
 {
 	EVP_PKEY_METHOD *orig_meth, *new_meth;
 
@@ -636,6 +821,54 @@ static EVP_PKEY_METHOD *pkcs11_pkey_method_ec()
 	return new_meth;
 }
 
+# if OPENSSL_VERSION_NUMBER >= 0x30000000L
+static EVP_PKEY_METHOD *pkcs11_pkey_method_ed25519(void)
+{
+	EVP_PKEY_METHOD *orig_meth, *new_meth;
+
+	orig_meth = (EVP_PKEY_METHOD *)EVP_PKEY_meth_find(EVP_PKEY_ED25519);
+
+	/* The digestsign() method is used to generate a signature in a one-shot mode */
+	EVP_PKEY_meth_get_digestsign(orig_meth, &orig_pkey_ed25519_digestsign);
+	
+	/* Don't assume any digest related defaults */
+	new_meth = EVP_PKEY_meth_new(EVP_PKEY_ED25519, EVP_PKEY_FLAG_SIGCTX_CUSTOM);
+
+	/* Duplicate the original method */
+	EVP_PKEY_meth_copy(new_meth, orig_meth);
+
+	/* Override selected ED25519 method callbacks with PKCS#11 implementations */
+	EVP_PKEY_meth_set_sign(new_meth, NULL, pkcs11_eddsa_pmeth_sign);
+	EVP_PKEY_meth_set_digestsign(new_meth, pkcs11_pkey_ed25519_digestsign);
+	EVP_PKEY_meth_set_ctrl(new_meth, pkcs11_eddsa_pmeth_ctrl, NULL);
+
+	return new_meth;
+}
+
+static EVP_PKEY_METHOD *pkcs11_pkey_method_ed448(void)
+{
+	EVP_PKEY_METHOD *orig_meth, *new_meth;
+
+	orig_meth = (EVP_PKEY_METHOD *)EVP_PKEY_meth_find(EVP_PKEY_ED448);
+	
+	/* The digestsign() method is used to generate a signature in a one-shot mode */
+	EVP_PKEY_meth_get_digestsign(orig_meth, &orig_pkey_ed448_digestsign);
+
+	/* Don't assume any digest related defaults */
+	new_meth = EVP_PKEY_meth_new(EVP_PKEY_ED448, EVP_PKEY_FLAG_SIGCTX_CUSTOM);
+
+	/* Duplicate the original method */
+	EVP_PKEY_meth_copy(new_meth, orig_meth);
+
+	/* Override selected ED448 method callbacks with PKCS#11 implementations */
+	EVP_PKEY_meth_set_sign(new_meth, NULL, pkcs11_eddsa_pmeth_sign);
+	EVP_PKEY_meth_set_digestsign(new_meth, pkcs11_pkey_ed448_digestsign);
+	EVP_PKEY_meth_set_ctrl(new_meth, pkcs11_eddsa_pmeth_ctrl, NULL);
+
+	return new_meth;
+}
+# endif /* OPENSSL_VERSION_NUMBER >= 0x30000000L */
+
 #endif /* OPENSSL_NO_EC */
 
 int PKCS11_pkey_meths(ENGINE *e, EVP_PKEY_METHOD **pmeth,
@@ -645,12 +878,20 @@ int PKCS11_pkey_meths(ENGINE *e, EVP_PKEY_METHOD **pmeth,
 		EVP_PKEY_RSA,
 #ifndef OPENSSL_NO_EC
 		EVP_PKEY_EC,
+# if OPENSSL_VERSION_NUMBER >= 0x30000000L
+		EVP_PKEY_ED25519,
+		EVP_PKEY_ED448,
+# endif /* OPENSSL_VERSION_NUMBER >= 0x30000000L */
 #endif /* OPENSSL_NO_EC */
 		0
 	};
 	static EVP_PKEY_METHOD *pkey_method_rsa = NULL;
 #ifndef OPENSSL_NO_EC
 	static EVP_PKEY_METHOD *pkey_method_ec = NULL;
+# if OPENSSL_VERSION_NUMBER >= 0x30000000L
+	static EVP_PKEY_METHOD *pkey_method_ed448 = NULL;
+	static EVP_PKEY_METHOD *pkey_method_ed25519 = NULL;
+# endif /* OPENSSL_VERSION_NUMBER >= 0x30000000L */
 #endif /* OPENSSL_NO_EC */
 
 	(void)e; /* squash the unused parameter warning */
@@ -678,10 +919,28 @@ int PKCS11_pkey_meths(ENGINE *e, EVP_PKEY_METHOD **pmeth,
 			return 0;
 		*pmeth = pkey_method_ec;
 		return 1; /* success */
+# if OPENSSL_VERSION_NUMBER >= 0x30000000L
+	case EVP_PKEY_ED448:
+		if (!pkey_method_ed448)
+			pkey_method_ed448 = pkcs11_pkey_method_ed448();
+		if (!pkey_method_ed448)
+			return 0;
+		*pmeth = pkey_method_ed448;
+		return 1; /* success */
+	case EVP_PKEY_ED25519:
+		if (!pkey_method_ed25519)
+			pkey_method_ed25519 = pkcs11_pkey_method_ed25519();
+		if (!pkey_method_ed25519)
+			return 0;
+		*pmeth = pkey_method_ed25519;
+		return 1; /* success */
+# endif /* OPENSSL_VERSION_NUMBER >= 0x30000000L */
 #endif /* OPENSSL_NO_EC */
 	}
 	*pmeth = NULL;
 	return 0;
 }
 
+#endif /* OPENSSL_VERSION_NUMBER < 0x40000000L */
+
 /* vim: set noexpandtab: */
