@@ -150,6 +150,9 @@ typedef struct {
 /* Utility functions                                                          */
 /******************************************************************************/
 
+#ifdef __GNUC__
+	__attribute__((format(printf, 6, 7)))
+#endif
 static void PROVIDER_CTX_log(PROVIDER_CTX *prov_ctx, int level, int reason, int line, const char *file, const char *format, ...)
 {
 	va_list args;
@@ -690,37 +693,8 @@ static int store_load(void *ctx, OSSL_CALLBACK *object_cb, void *object_cbarg,
 			"No custom UI method provided, using the default UI method.\n");
 	}
 
-	/* try fetching a private key */
-	if (store_ctx->types_tried == 0) {
-		store_ctx->types_tried++;
-		if (store_ctx->expected_type == 0 || store_ctx->expected_type == OSSL_STORE_INFO_PKEY) {
-			EVP_PKEY *key = UTIL_CTX_get_privkey_from_uri(store_ctx->prov_ctx->util_ctx,
-				store_ctx->uri, ui_method, ui_data);
-
-			UTIL_CTX_set_ui_method(store_ctx->prov_ctx->util_ctx, ui_method, NULL);
-			if (key != NULL) {
-				/* Workaround for EVP_PKEY without key management, needed since
-				 * ossl_store_handle_load_result() doesn't support this case. */
-				cbdata->v = OSSL_STORE_INFO_new_PKEY(key);
-				return 1;
-			}
-		}
-	}
-	/* try fetching a public key */
-	if (store_ctx->types_tried == 1) {
-		store_ctx->types_tried++;
-		if (store_ctx->expected_type == 0 || store_ctx->expected_type == OSSL_STORE_INFO_PUBKEY) {
-			EVP_PKEY *key = UTIL_CTX_get_pubkey_from_uri(store_ctx->prov_ctx->util_ctx,
-				store_ctx->uri, ui_method, ui_data);
-
-			if (key != NULL) {
-				cbdata->v = OSSL_STORE_INFO_new_PUBKEY(key);
-				return 1;
-			}
-		}
-	}
 	/* try fetching a certificate  */
-	if (store_ctx->types_tried == 2) {
+	if (store_ctx->types_tried == 0) {
 		store_ctx->types_tried++;
 		if (store_ctx->expected_type == 0 || store_ctx->expected_type ==  OSSL_STORE_INFO_CERT) {
 			X509 *cert = UTIL_CTX_get_cert_from_uri(store_ctx->prov_ctx->util_ctx,
@@ -735,13 +709,13 @@ static int store_load(void *ctx, OSSL_CALLBACK *object_cb, void *object_cbarg,
 				int len = i2d_X509(cert, NULL);
 
 				if (len < 0) {
-					PROVIDER_CTX_log(store_ctx->prov_ctx, LOG_ERR, 4, OPENSSL_LINE, OPENSSL_FUNC, store_ctx->uri);
+					PROVIDER_CTX_log(store_ctx->prov_ctx, LOG_ERR, 4, OPENSSL_LINE, OPENSSL_FUNC, "%s", store_ctx->uri);
 					X509_free(cert);
 					return 0;
 				}
 				tmp = data = OPENSSL_malloc((size_t)len);
 				if (!tmp) {
-					PROVIDER_CTX_log(store_ctx->prov_ctx, LOG_ERR, 1, OPENSSL_LINE, OPENSSL_FUNC, store_ctx->uri);
+					PROVIDER_CTX_log(store_ctx->prov_ctx, LOG_ERR, 1, OPENSSL_LINE, OPENSSL_FUNC, "%s", store_ctx->uri);
 					X509_free(cert);
 					return 0;
 				}
@@ -755,7 +729,7 @@ static int store_load(void *ctx, OSSL_CALLBACK *object_cb, void *object_cbarg,
 
 				if (!object_cb(params, object_cbarg)) {
 					/* callback failed */
-					PROVIDER_CTX_log(store_ctx->prov_ctx, LOG_ERR, 5, OPENSSL_LINE, OPENSSL_FUNC, store_ctx->uri);
+					PROVIDER_CTX_log(store_ctx->prov_ctx, LOG_ERR, 5, OPENSSL_LINE, OPENSSL_FUNC, "%s", store_ctx->uri);
 					OPENSSL_free(data);
 					return 0;
 				}
@@ -764,12 +738,46 @@ static int store_load(void *ctx, OSSL_CALLBACK *object_cb, void *object_cbarg,
 			}
 		}
 	}
+	/* try fetching a public key */
+	if (store_ctx->types_tried == 1) {
+		store_ctx->types_tried++;
+		if (store_ctx->expected_type == 0 || store_ctx->expected_type == OSSL_STORE_INFO_PUBKEY) {
+			EVP_PKEY *key = UTIL_CTX_get_pubkey_from_uri(store_ctx->prov_ctx->util_ctx,
+				store_ctx->uri, ui_method, ui_data);
+
+			if (key != NULL) {
+				/* Workaround for EVP_PKEY without key management, needed since
+				 * ossl_store_handle_load_result() doesn't support this case. */
+				cbdata->v = OSSL_STORE_INFO_new_PUBKEY(key);
+				return 1;
+			}
+		}
+	}
+	/* try fetching a private key */
+	if (store_ctx->types_tried == 2) {
+		store_ctx->types_tried++;
+		if (store_ctx->expected_type == 0 || store_ctx->expected_type == OSSL_STORE_INFO_PKEY) {
+			EVP_PKEY *key = UTIL_CTX_get_privkey_from_uri(store_ctx->prov_ctx->util_ctx,
+				store_ctx->uri, ui_method, ui_data);
+
+			UTIL_CTX_set_ui_method(store_ctx->prov_ctx->util_ctx, ui_method, NULL);
+			if (key != NULL) {
+				/* Workaround for EVP_PKEY without key management, needed since
+				 * ossl_store_handle_load_result() doesn't support this case. */
+				cbdata->v = OSSL_STORE_INFO_new_PKEY(key);
+				return 1;
+			}
+		}
+	}
 	return 0;
 }
 
 /*
  * Indicates whether all expected objects from the URI have been processed.
- * The expected sequence is: a private key, a public key, and a certificate.
+ * The expected sequence is:
+ * 0 - OSSL_STORE_INFO_CERT   - X.509 certificate (X509 *)
+ * 1 - OSSL_STORE_INFO_PUBKEY - public key (EVP_PKEY *)
+ * 2 - OSSL_STORE_INFO_PKEY   - private key (EVP_PKEY *)
  * Once the counter reaches 3, all objects have been handled, making further
  * loading attempts unnecessary.
  */
