@@ -33,6 +33,8 @@
 #include <openssl/pem.h>
 #include <openssl/err.h>
 
+#ifndef OPENSSL_NO_ENGINE
+
 static void usage(char *argv[])
 {
 	fprintf(stderr, "%s [private key URL] [module] [conf]\n", argv[0]);
@@ -58,7 +60,9 @@ int main(int argc, char *argv[])
 {
 	ENGINE *engine = NULL;
 	EVP_PKEY *pkey = NULL;
+#ifndef OPENSSL_NO_EC
 	EC_KEY *ec = NULL, *ec_dup = NULL;
+#endif /* OPENSSL_NO_EC */
 
 	const char *module, *efile, *privkey;
 
@@ -138,6 +142,7 @@ int main(int argc, char *argv[])
 	case EVP_PKEY_RSA:
 		/* TODO */
 		break;
+#ifndef OPENSSL_NO_EC
 	case EVP_PKEY_EC:
 		ec = EVP_PKEY_get1_EC_KEY(pkey);
 		if (ec == NULL) {
@@ -157,6 +162,7 @@ int main(int argc, char *argv[])
 		EC_KEY_free(ec);
 		EC_KEY_free(ec_dup);
 		break;
+#endif /* OPENSSL_NO_EC */
 	}
 
 	EVP_PKEY_free(pkey);
@@ -181,4 +187,15 @@ end:
 	return ret;
 }
 
+#else /* OPENSSL_NO_ENGINE */
+
+#include <stdio.h>
+
+int main() {
+	fprintf(stderr, "Skipped: ENGINE support not available\n");
+	return 77;
+}
+
+#endif /* OPENSSL_NO_ENGINE */
+
 /* vim: set noexpandtab: */
