@@ -44,6 +44,8 @@
 #include <openssl/conf.h>
 #include <openssl/ui.h>
 
+#ifndef OPENSSL_NO_ENGINE
+
 /* UI method that's only used to fail if get_pin inside engine_pkcs11
  * has failed to pick up in a PIN sent in with ENGINE_ctrl_cmd_string */
 static UI_METHOD *ui_detect_failed_ctrl = NULL;
@@ -105,7 +107,7 @@ static int ui_write(UI *ui, UI_STRING *uis)
 	return UI_method_get_writer(UI_OpenSSL())(ui, uis);
 }
 
-static void setup_ui()
+static void setup_ui(void)
 {
 	UI_METHOD *default_method = UI_OpenSSL();
 
@@ -319,4 +321,15 @@ int main(int argc, char **argv)
 	return 0;
 }
 
+#else /* OPENSSL_NO_ENGINE */
+
+#include <stdio.h>
+
+int main() {
+	fprintf(stderr, "Skipped: ENGINE support not available\n");
+	return 77;
+}
+
+#endif /* OPENSSL_NO_ENGINE */
+
 /* vim: set noexpandtab: */
