@@ -66,6 +66,8 @@
 #define RANDOM_SIZE 20
 #define MAX_SIGSIZE 1024
 
+#ifndef OPENSSL_NO_ENGINE
+
 static int do_wait(pid_t pids[], int num)
 {
 	int i;
@@ -315,4 +317,15 @@ failed:
 	return rv;
 }
 
+#else /* OPENSSL_NO_ENGINE */
+
+#include <stdio.h>
+
+int main() {
+	fprintf(stderr, "Skipped: ENGINE support not available\n");
+	return 77;
+}
+
+#endif /* OPENSSL_NO_ENGINE */
+
 /* vim: set noexpandtab: */
