@@ -40,6 +40,8 @@
 #include <openssl/engine.h>
 #include <openssl/conf.h>
 
+#ifndef OPENSSL_NO_ENGINE
+
 static void display_openssl_errors(int l)
 {
 	const char *file;
@@ -270,4 +272,15 @@ int main(int argc, char **argv)
 	return 0;
 }
 
+#else /* OPENSSL_NO_ENGINE */
+
+#include <stdio.h>
+
+int main() {
+	fprintf(stderr, "Skipped: ENGINE support not available\n");
+	return 77;
+}
+
+#endif /* OPENSSL_NO_ENGINE */
+
 /* vim: set noexpandtab: */
